#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "resource.h"
#include "resource\BGFont.h"
//#include "resource\BitMapFont.h"

#include <stdlib.h>

//---------------------------------------------------------------------------
//^C̃TCY
#define BG_CX               8
#define BG_CY               8

//Xe[W̃TCY
#define STAGE_CX            (10+2)	//+2͊O̕ǔzup
#define STAGE_CY            (18+2)

//Xe[WŎgϐ
#define STAGE_EMPTY         0
#define STAGE_BLOCK         1
#define STAGE_WALL          2

//ubNXg
#define BLOCK_LIST_TYPE     7
#define BLOCK_LIST_CX       4
#define BLOCK_LIST_CY       4

//\pBG^C
#define BG_TILE_EMPTY       0
#define BG_TILE_BLOCK       1
#define BG_TILE_WALL        2

//Q[̐ݒ
#define BLOCK_SPEED         0x3000
#define BLOCK_APPEAR_X      4		//ubN̏\ʒu
#define BLOCK_APPEAR_Y      0
#define BLOCK_NEXT_X        13		//mdwsubN̕\ʒu
#define BLOCK_NEXT_Y        4
#define SCORE_X             13
#define SCORE_Y             1

//Q[̐is
#define GAME_STATE_MAKE     1	//ubN̍쐬
#define GAME_STATE_CHECK    2	//ubN̈ړ`FbN
#define GAME_STATE_INPUT    3	//L[̓
#define GAME_STATE_MOVE     4	//ubN̈ړ
#define GAME_STATE_COPY     5	//ubNXe[WɃRs[
#define GAME_STATE_DELETE   6	//ubN̏
#define GAME_STATE_END      7	//Q[̏I
//---------------------------------------------------------------------------
//egXp̃ubNXg
u8 BlockList[BLOCK_LIST_TYPE][BLOCK_LIST_CX][BLOCK_LIST_CY] =
{
	{
		{0,1,0,0},
		{0,1,0,0},
		{0,1,0,0},
		{0,1,0,0},
	},
	{
		{0,0,0,0},
		{0,1,1,0},
		{0,1,0,0},
		{0,1,0,0},
	},
	{
		{0,0,0,0},
		{0,1,1,0},
		{0,0,1,0},
		{0,0,1,0},
	},
	{
		{0,0,1,0},
		{0,1,1,0},
		{0,1,0,0},
		{0,0,0,0},
	},
	{
		{0,1,0,0},
		{0,1,1,0},
		{0,0,1,0},
		{0,0,0,0},
	},
	{
		{0,0,0,0},
		{0,1,0,0},
		{1,1,1,0},
		{0,0,0,0},
	},
	{
		{0,0,0,0},
		{0,1,1,0},
		{0,1,1,0},
		{0,0,0,0},
	},
};
//---------------------------------------------------------------------------
typedef struct {
	u8  x;
	u8  y;
	u8  key;		//ړp
	u16 score;		//XRA
	u16 BlockSpeed;	//Xs[h
	u8  BlockNext;
	u8  Block[BLOCK_LIST_CX][BLOCK_LIST_CY];
} ST_PLAYER;

ST_PLAYER PLAYER;

//---------------------------------------------------------------------------
u8 BG2Chr;
u8 BG2Map;
u8 BG2Col;

u8 BG3Chr;
u8 BG3Map;
u8 BG3Col;

u8 GameStateFlag;				//Q[̐is
u8 BlockFallFlag;				//ubN̗tO
u8 Stage[STAGE_CX][STAGE_CY];
//---------------------------------------------------------------------------
void Init();
void InitMap();
void InitPlayer();

void GameStateMake();
void GameStateCheck();
void GameStateInput();
void GameStateMove();
void GameStateCopy();
void GameStateDelete();

void DrawStageAll();
void DrawStage(const u8 x, const u8 y, const u8 chr);
void DrawBlock(const u8 argX, const u8 argY, const bool argDelFlag);
void DrawNextBlock();
void DrawBGStage(const u8 bgX, const u8 bgY, const u8 tileNo);
void DrawBGGameOver();
void DrawBGScore();

void Input();

void SetBlockNext();
void TurnBlock();

bool ChkMoveBlock(s8 argX, s8 argY, u8 argBlock[BLOCK_LIST_CX][BLOCK_LIST_CY]);
bool ChkTurnBlock(u8 argX, u8 argY);

void IRQUserSetup();
void IRQUserHandler();

void TimerStart();
void TimerStop();
void TimerReset();
//---------------------------------------------------------------------------
int main(void)
{
//	WaitKey(KEY_A);	//srand

	Init();
	InitMap();
	InitPlayer();

	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	TimerStart();

	for(;;)
	{
		WaitForVsync();

		switch(GameStateFlag)
		{
			case GAME_STATE_MAKE:   GameStateMake();   break;
			case GAME_STATE_CHECK:  GameStateCheck();  break;
			case GAME_STATE_INPUT:  GameStateInput();  break;
			case GAME_STATE_MOVE:   GameStateMove();   break;
			case GAME_STATE_COPY:   GameStateCopy();   break;
			case GAME_STATE_DELETE: GameStateDelete(); break;
			case GAME_STATE_END:    goto End;          break;
			default: DebugMessage((u8*)"Err: Main GameStateFlag"); break;
		}
	}

End:

	TimerStop();
	DrawBGGameOver();

	for(;;){}
	return 0;
}
//---------------------------------------------------------------------------
void Init()
{
	//EFCgRg[
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;

	//^C}[荞
	IRQUserSetup();

	//L[
	KeyInit();

	//BG
	//BG2 \p
	//BG3 Q[p
	BG2Chr = 1;
	BG2Map = 0;
	BG2Col = 0;

	BG3Chr = 2;
	BG3Map = 1;
	BG3Col = 1;

	REG_BG2CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG2Chr) | BG_MAPBASE(BG2Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG3Chr) | BG_MAPBASE(BG3Map) | 0);

	BGSetPal16  (BG2Col, (u16*)&BGFont_palette);
	BGSetChrData(BG2Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);

	BGSetPal16  (BG3Col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG3Chr, (u16*)&game_bg_img_bin, 2048);
	BGSetPalMap (BG3Map, 2048, 1);
}
//---------------------------------------------------------------------------
void InitMap()
{
	//Map̏
	u8 x,y;
	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			Stage[x][y] = STAGE_EMPTY;
		}
	}

	//Oǂ̍쐬
	for(x=0; x<STAGE_CX; x++)
	{
		Stage[x][STAGE_CY-1] = STAGE_WALL;
	}
	for(y=0; y<STAGE_CY; y++)
	{
		Stage[0         ][y] = STAGE_WALL;
		Stage[STAGE_CX-1][y] = STAGE_WALL;
	}

	//Map\
	DrawStageAll();
}
//---------------------------------------------------------------------------
void InitPlayer()
{
	GameStateFlag     = GAME_STATE_MAKE;
	BlockFallFlag     = FALSE;

	PLAYER.score      = 0;
	PLAYER.BlockNext  = rand() % BLOCK_LIST_TYPE;
	PLAYER.BlockSpeed = BLOCK_SPEED;
}
//---------------------------------------------------------------------------
void DrawStageAll()
{
	u8 x,y;
	
	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			DrawStage(x,y,Stage[x][y]);
		}
	}
}
//---------------------------------------------------------------------------
void DrawStage(const u8 x, const u8 y, const u8 chr)
{
	u8 tileNo = 0;
	switch(chr)
	{
		case STAGE_EMPTY: tileNo = BG_TILE_EMPTY; break;
		case STAGE_BLOCK: tileNo = BG_TILE_BLOCK; break;
		case STAGE_WALL : tileNo = BG_TILE_WALL ; break;
		default: DebugMessage((u8*)"Err: DrawStage Tile"); break;
	}

	DrawBGStage(x, y, tileNo);
}
//---------------------------------------------------------------------------
void DrawBGStage(const u8 bgX, const u8 bgY, const u8 tileNo)
{
	BGDrawChar(BG3Map, bgX, bgY, BG3Col, tileNo);
}
//---------------------------------------------------------------------------
void DrawBlock(const u8 argX, const u8 argY, const bool argDisplayFlag)
{
	u8 x, y;

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			if(PLAYER.Block[x][y] == STAGE_BLOCK)
			{
				if(argDisplayFlag == TRUE)
				{
					DrawStage(x+argX, y+argY, STAGE_BLOCK);
				}
				else
				{
					DrawStage(x+argX, y+argY, STAGE_EMPTY);
				}
			}
		}
	}
}
//---------------------------------------------------------------------------
void DrawBGGameOver()
{
	char msg[40];
	bprintf(msg, "Game Over");
	BGDrawStr(BG2Map, 13, 18, (u8*)msg);
}
//---------------------------------------------------------------------------
void IRQUserSetup()
{
	REG_IME = IRQ_MASTER_OFF;

	REG_INTERUPT  = (u32)IRQUserHandler;
	REG_IE       |= IRQ_BIT_TIMER0;

	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	REG_IME  = IRQ_MASTER_OFF;
	u16 Flag = REG_IF;

	if(Flag & IRQ_BIT_TIMER0)
	{
		//ubN̗tOnmɂ܂
		BlockFallFlag = TRUE;
	}

	REG_IF  = Flag;
	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void TimerStart()
{
	REG_TM0D   = PLAYER.BlockSpeed;
	REG_TM0CNT = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void TimerStop()
{
	REG_TM0D   = 0x0000;
	REG_TM0CNT = TM_DISABLE;
}
//---------------------------------------------------------------------------
void TimerReset()
{
	REG_TM0D      = PLAYER.BlockSpeed;
	BlockFallFlag = FALSE;
}
//---------------------------------------------------------------------------
//ubN̈ړ`FbN(TRUE: ړł FALSE: ړłȂ)
bool ChkMoveBlock(s8 argX, s8 argY, u8 argBlock[BLOCK_LIST_CX][BLOCK_LIST_CY])
{
	u8 x, y;

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			if(argBlock[x][y] > 0)
			{
				if(Stage[x+argX][y+argY] > 0)
				{
					return FALSE;
				}
			}
		}
	}

	return TRUE;
}
//---------------------------------------------------------------------------
void SetBlockNext()
{
	u8 x, y;

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			PLAYER.Block[x][y] = BlockList[PLAYER.BlockNext][y][x];
		}
	}
}
//---------------------------------------------------------------------------
void SetBlockToStage()
{
	u8 x, y;

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			if(PLAYER.Block[x][y] > 0)
			{
				Stage[x+PLAYER.x][y+PLAYER.y] = PLAYER.Block[x][y];
			}
		}
	}
}
//---------------------------------------------------------------------------
void Input()
{
	u8 Key = KeyInput();

	if(Key == 0x00)
	{
		return;
	}

	PLAYER.key = 0x00;

	if(Key & KEY_DOWN ) PLAYER.key = KEY_DOWN;
	if(Key & KEY_LEFT ) PLAYER.key = KEY_LEFT;
	if(Key & KEY_RIGHT) PLAYER.key = KEY_RIGHT;
	if(Key & KEY_A    ) PLAYER.key = KEY_A;

	GameStateFlag = GAME_STATE_CHECK;
}
//---------------------------------------------------------------------------
void GameStateMake()
{
	SetBlockNext();

	PLAYER.BlockNext = rand() % BLOCK_LIST_TYPE;
	PLAYER.x         = BLOCK_APPEAR_X;
	PLAYER.y         = BLOCK_APPEAR_Y;
	PLAYER.key       = 0x00;

	DrawNextBlock();
	DrawBlock(PLAYER.x, PLAYER.y, TRUE);
	DrawBGScore();

	TimerReset();
	GameStateFlag = GAME_STATE_INPUT;
}
//---------------------------------------------------------------------------
void GameStateInput()
{
	//tÕ`FbN
	if(BlockFallFlag == TRUE)
	{
		PLAYER.key    = KEY_DOWN;
		GameStateFlag = GAME_STATE_CHECK;
		return;
	}
	
	//L[
	Input();
}
//---------------------------------------------------------------------------
void GameStateCheck()
{
	switch(PLAYER.key)
	{
	case KEY_DOWN:
		if(ChkMoveBlock(PLAYER.x, PLAYER.y + 1, PLAYER.Block) == TRUE)
		{
			GameStateFlag = GAME_STATE_MOVE;
		}
		else if(PLAYER.x == BLOCK_APPEAR_X && PLAYER.y == BLOCK_APPEAR_Y)
		{
			GameStateFlag = GAME_STATE_END;
		}
		else
		{
			GameStateFlag = GAME_STATE_COPY;
		}
		break;

	case KEY_RIGHT:
		if(ChkMoveBlock(PLAYER.x+1, PLAYER.y, PLAYER.Block) == TRUE)
		{
			GameStateFlag = GAME_STATE_MOVE;
		}
		else
		{
			GameStateFlag = GAME_STATE_INPUT;
		}
		break;

	case KEY_LEFT:
		if(ChkMoveBlock(PLAYER.x-1, PLAYER.y, PLAYER.Block) == TRUE)
		{
			GameStateFlag = GAME_STATE_MOVE;
		}
		else
		{
			GameStateFlag = GAME_STATE_INPUT;
		}
		break;

	case KEY_A:
		if(ChkTurnBlock(PLAYER.x, PLAYER.y) == TRUE)
		{
			GameStateFlag = GAME_STATE_MOVE;
		}
		else
		{
			GameStateFlag = GAME_STATE_INPUT;
		}
		break;

	default:
		DebugMessage((u8*)"Err: GameStateCheck PLAYER.key");
		break;
	}
}
//---------------------------------------------------------------------------
void GameStateMove()
{
	DrawBlock(PLAYER.x, PLAYER.y, FALSE);

	if(PLAYER.key & KEY_DOWN ){ PLAYER.y++; TimerReset(); }
	if(PLAYER.key & KEY_LEFT ){ PLAYER.x--; }
	if(PLAYER.key & KEY_RIGHT){ PLAYER.x++; }
	if(PLAYER.key & KEY_A    ){ TurnBlock();}

	DrawBlock(PLAYER.x, PLAYER.y, TRUE);

	GameStateFlag = GAME_STATE_INPUT;
}
//---------------------------------------------------------------------------
void GameStateCopy()
{
	SetBlockToStage();
	GameStateFlag = GAME_STATE_DELETE;
}
//---------------------------------------------------------------------------
void GameStateDelete()
{
	u8 x,y,x2,y2;
	u8 DeleteLineFlag;
	u8 DeleteLineCnt = 0;

	for(y=1; y<STAGE_CY-1; y++)
	{
		//s͍폜\`FbN܂
		DeleteLineFlag = TRUE;
		for(x=1; x<STAGE_CX-1; x++)
		{
			if(Stage[x][y] == 0) DeleteLineFlag = FALSE;
		}

		if(DeleteLineFlag == TRUE)
		{
			//s̍폜JEgCNg
			DeleteLineCnt++;
			
			//s̍폜
			for(x=1; x<STAGE_CX-1; x++)
			{
				Stage[x][y] = 0;
			}
			//폜s̋󔒂l߂܂
			for(y2=y; y2>=1; y2--)
			{
				for(x2=1; x2<=STAGE_CX-1; x2++)
				{
					Stage[x2][y2] = Stage[x2][y2-1];
				}
			}
		}
	}

	PLAYER.score += DeleteLineCnt * 100;

	DrawStageAll();
	GameStateFlag = GAME_STATE_MAKE;
}
//---------------------------------------------------------------------------
bool ChkTurnBlock(u8 argX, u8 argY)
{
	u8 tmpBlcok[BLOCK_LIST_CX][BLOCK_LIST_CY];
	u8 x, y;

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			tmpBlcok[x][y] = PLAYER.Block[(BLOCK_LIST_CY-1)-y][x];
		}
	}

	return ChkMoveBlock(argX, argY, tmpBlcok);
}
//---------------------------------------------------------------------------
void TurnBlock()
{
	u8 tmpBlcok[BLOCK_LIST_CX][BLOCK_LIST_CY];
	u8 x, y;

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			tmpBlcok[x][y] = PLAYER.Block[(BLOCK_LIST_CY-1)-y][x];
		}
	}

	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			PLAYER.Block[x][y] = tmpBlcok[x][y];
		}
	}
}
//---------------------------------------------------------------------------
void DrawNextBlock()
{
	u8 x,y;
	
	for(y=0; y<BLOCK_LIST_CY; y++)
	{
		for(x=0; x<BLOCK_LIST_CX; x++)
		{
			DrawStage(BLOCK_NEXT_X + x, BLOCK_NEXT_Y + y, BlockList[PLAYER.BlockNext][y][x]);
		}
	}
}
//---------------------------------------------------------------------------
void DrawBGScore()
{
	char score[40];
	bprintf(score, "%6d", PLAYER.score);
	BGDrawStr(BG2Map, SCORE_X, SCORE_Y, (u8*)score);
}
